@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2001,2011-2012 LAAS/CNRS
+ * Copyright (c) 2001,2011-2012,2020 LAAS/CNRS
  * All rights reserved.
  *
  * Redistribution and use  in source  and binary  forms,  with or without
@@ -55,8 +55,8 @@ elTclCompletion(EditLine *el, int ch)
    const LineInfo *linfo;
    Tcl_Obj *cmd[2], *buffer, *arg, **matches, *cmdLine;
    Tcl_Channel outChannel;
-   int i,j,k, start, end, count, length0, length1, max, oldLen;
-   char *string0, *string1, *old;
+   int i,j,k, start, end, count, length0, length1, max;
+   char *string0, *string1;
    int again, ncols, nitems;
    char c;
 
@@ -96,6 +96,7 @@ elTclCompletion(EditLine *el, int ch)
    }
 
    Tcl_DecrRefCount(cmd[1]);
+   Tcl_DecrRefCount(cmdLine);
 
    /* get result and display */
    buffer = Tcl_GetObjResult(iinfo->interp);
@@ -103,9 +104,8 @@ elTclCompletion(EditLine *el, int ch)
    Tcl_GetIntFromObj(iinfo->interp, matches[0], &start);
    Tcl_GetIntFromObj(iinfo->interp, matches[1], &end);
 
-   if (count == 2) {
+   if (count < 3) {
       /* no match */
-      Tcl_DecrRefCount(cmdLine);
       return CC_ERROR;
    }
 
@@ -117,32 +117,9 @@ elTclCompletion(EditLine *el, int ch)
       el_insertstr(el, Tcl_GetStringFromObj(arg, NULL));
       Tcl_ListObjIndex(iinfo->interp, matches[2], 1, &arg);
       el_insertstr(el, Tcl_GetStringFromObj(arg, NULL));
-      Tcl_DecrRefCount(cmdLine);
       return CC_REFRESH;
    }
 
-   if (count-2 > iinfo->completionQueryItems && iinfo->completionQueryItems) {
-      /* ask user */
-      printf("\nDisplay all %d possibilit%s? [y/n] ",
-	     count-2, count>3?"ies":"y");
-      fflush(stdout);
-
-      if (el_getc(iinfo->el, &c) <= 0) {
-	 fputc('\n', stdout);
-	 Tcl_DecrRefCount(cmdLine);
-	 return CC_REDISPLAY;
-      }
-
-      if (c != 'y' && c != 'Y') {
-	 fputc(c, stdout);
-	 fputc('\n', stdout);
-	 Tcl_DecrRefCount(cmdLine);
-	 return CC_REDISPLAY;
-      }
-
-      fputc(c, stdout);
-   }
-
    /* find the smallest common part in every matches */
    Tcl_ListObjIndex(iinfo->interp, matches[2], 0, &cmd[0]);
    string0 = Tcl_GetStringFromObj(cmd[0], &length0);
@@ -162,17 +139,38 @@ elTclCompletion(EditLine *el, int ch)
 	 }
       }
    } while (again);
+   end++;
 
-   old = Tcl_GetStringFromObj(cmdLine, &oldLen);
-   if (end+start >= oldLen) {
-      c = string0[end+1];
-      string0[end+1] = 0;
-      el_insertstr(el, string0);
-      string0[end+1] = c;
-   } else {
-      el_insertstr(el, old + start);
+   /* insert common part */
+   if (end > 0) {
+     char str[end+1];
+     strncpy(str, string0, end);
+     str[end] = 0;
+     el_insertstr(el, str);
+   }
+
+   /* ask user if too many items are to be displayed */
+   if (count-2 > iinfo->completionQueryItems && iinfo->completionQueryItems) {
+     if (end > linfo->cursor - linfo->buffer - start)
+       return CC_REDISPLAY;
+
+      printf("\nDisplay all %d possibilit%s? [y/n] ",
+	     count-2, count>3?"ies":"y");
+      fflush(stdout);
+
+      if (el_getc(iinfo->el, &c) <= 0) {
+	 fputc('\n', stdout);
+	 return CC_REDISPLAY;
+      }
+
+      if (c != 'y' && c != 'Y') {
+	 fputc(c, stdout);
+	 fputc('\n', stdout);
+	 return CC_REDISPLAY;
+      }
+
+      fputc(c, stdout);
    }
-   Tcl_DecrRefCount(cmdLine);
 
    /* find the biggest match (for multicol display) */
    max = 0;
