@@ -1,8 +1,6 @@
-/*	$LAAS$ */
-
-/* 
- * Copyright (c) 2001 LAAS/CNRS                       --  Sun Oct 14 2001
- * All rights reserved.                                    Anthony Mallet
+/*
+ * Copyright (c) 2001,2010-2012,2017 LAAS/CNRS
+ * All rights reserved.
  *
  * Redistribution and use  in source  and binary  forms,  with or without
  * modification, are permitted provided that the following conditions are
@@ -27,11 +25,13 @@
  * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
  * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
  * DAMAGE.
+ *
+ *                                      Anthony Mallet on Sun Oct 14 2001
  */
-#include "config.h"
-__RCSID("$LAAS$");
+#include "elconfig.h"
 
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 #include <sys/ioctl.h>
@@ -74,7 +74,8 @@ elTclGets(ClientData data, Tcl_Interp *interp,
       code = TCL_ERROR;
    } else {
       Tcl_SetObjResult(interp, Tcl_NewStringObj(line, length-1));
-      history(iinfo->askaHistory, &ev, H_ENTER, line);
+      if (length > 1)
+        history(iinfo->askaHistory, &ev, H_ENTER, line);
    }
 
    /* set editline back to normal state */
@@ -124,28 +125,121 @@ int
 elTclHistory(ClientData data, Tcl_Interp *interp,
 	     int objc, Tcl_Obj *const objv[])
 {
-   ElTclInterpInfo *iinfo = data;
-   char *string;
-   HistEvent ev;
+  enum hidx {
+    hidx_add, hidx_change, hidx_event, hidx_file, hidx_save, hidx_size,
+    hidx_clear
+  };
+  static const char *args[] = {
+    [hidx_add] = "add", [hidx_change] = "change", [hidx_event] = "event",
+    [hidx_file] = "file", [hidx_save] = "save", [hidx_size] = "size",
+    [hidx_clear] = "clear", NULL
+  };
 
-   if (objc != 3) {
-      Tcl_WrongNumArgs(interp, 1, objv, "add string");
-      return TCL_ERROR;
-   }
+  ElTclInterpInfo *iinfo = data;
+  char *string;
+  HistEvent ev;
+  int s;
+  int i = -1;
 
-   if (strcmp(Tcl_GetStringFromObj(objv[1], NULL), "add")) {
-      Tcl_WrongNumArgs(interp, 1, objv, "add string");
-      return TCL_ERROR;
-   }
-   
-   string = Tcl_GetStringFromObj(objv[2], NULL);
-   if (string == NULL) {
-      Tcl_WrongNumArgs(interp, 1, objv, "add string");
-      return TCL_ERROR;
-   }
+  if (objc > 1) {
+    s = Tcl_GetIndexFromObj(interp, objv[1], args, "subcommand", 0, &i);
+    if (s != TCL_OK) return s;
+  }
+  switch(i) {
+    case hidx_add:
+      if (objc != 3) {
+        Tcl_WrongNumArgs(interp, 2, objv, "string");
+        return TCL_ERROR;
+      }
+      string = Tcl_GetStringFromObj(objv[2], NULL);
+      history(iinfo->history, &ev, H_ENTER, string);
+      break;
 
-   history(iinfo->history, &ev, H_ENTER, string);
-   return TCL_OK;
+    case hidx_change:
+      if (objc > 4 || objc < 3) {
+        Tcl_WrongNumArgs(interp, 2, objv, "newValue ?event?");
+        return TCL_ERROR;
+      }
+      string = Tcl_GetStringFromObj(objv[2], NULL);
+      if (!string) abort();
+      if (objc > 3) {
+        if (Tcl_GetIntFromObj(interp, objv[3], &s) != TCL_OK)
+          return TCL_ERROR;
+        history(iinfo->history, &ev, s<=0?H_FIRST:H_LAST);
+        history(iinfo->history, &ev, H_SET, ev.num+s);
+        if (ev.num) {
+          Tcl_AppendResult(interp, ev.str, NULL);
+          return TCL_ERROR;
+        }
+      }
+      history(iinfo->history, &ev, H_CURR);
+      *(char *)ev.str = '\0'; /* sadly, one must hack into the internals */
+      history(iinfo->history, &ev, H_ADD, string);
+      break;
+
+    case hidx_event:
+      if (objc > 3) {
+        Tcl_WrongNumArgs(interp, 2, objv, "?event?");
+        return TCL_ERROR;
+      }
+      if (objc > 2) {
+        if (Tcl_GetIntFromObj(interp, objv[2], &s) != TCL_OK)
+          return TCL_ERROR;
+        history(iinfo->history, &ev, s<=0?H_FIRST:H_LAST);
+        history(iinfo->history, &ev, H_SET, ev.num+s);
+        if (ev.num) {
+          Tcl_AppendResult(interp, ev.str, NULL);
+          return TCL_ERROR;
+        }
+      }
+      history(iinfo->history, &ev, H_CURR);
+      string = Tcl_Alloc(strlen(ev.str)+1);
+      strcpy(string, ev.str);
+      while (string[strlen(string)-1] == '\n')
+        string[strlen(string)-1] = '\0';
+      Tcl_SetResult(interp, string, TCL_DYNAMIC);
+      break;
+
+    case hidx_file:
+      if (objc < 3) {
+        Tcl_SetObjResult(interp, Tcl_NewStringObj(iinfo->histFile, -1));
+        break;
+      }
+      if (iinfo->histFile) free((void *)iinfo->histFile);
+      Tcl_IncrRefCount(objv[2]);
+      iinfo->histFile = Tcl_FSGetNativePath(objv[2]);
+      if (iinfo->histFile) iinfo->histFile = strdup(iinfo->histFile);
+      if (iinfo->histFile && iinfo->histFile[0])
+        history(iinfo->history, &ev, H_LOAD, iinfo->histFile);
+      Tcl_DecrRefCount(objv[2]);
+      break;
+
+    case hidx_save:
+      if (iinfo->histFile && iinfo->histFile[0])
+        history(iinfo->history, &ev, H_SAVE, iinfo->histFile);
+      break;
+
+    case hidx_size:
+      if (objc < 3) {
+        Tcl_SetObjResult(interp, Tcl_NewIntObj(iinfo->histSize));
+        break;
+      }
+      if (Tcl_GetIntFromObj(interp, objv[2], &iinfo->histSize) != TCL_OK)
+        return TCL_ERROR;
+
+      history(iinfo->history, &ev, H_SETSIZE, iinfo->histSize);
+      break;
+
+    case hidx_clear:
+      if (objc != 2) {
+        Tcl_WrongNumArgs(interp, 2, objv, "");
+        return TCL_ERROR;
+      }
+      history(iinfo->history, &ev, H_CLEAR);
+      break;
+  }
+
+  return TCL_OK;
 }
 
 
@@ -157,24 +251,34 @@ elTclHistory(ClientData data, Tcl_Interp *interp,
  */
 
 int
-elTclEventLoop(EditLine *el, char *c)
+elTclEventLoop(EditLine *el, wchar_t *c)
 {
    ElTclInterpInfo *iinfo;
+   int mbSz;
+
    el_get(el, EL_CLIENTDATA, &iinfo);
-   
+
    /* process Tcl events until there's some input available */
    while (iinfo->preReadSz == 0 && !feof(stdin))
       Tcl_DoOneEvent(0);
 
    if (iinfo->preReadSz == 0 && feof(stdin)) {
       /* eof and no pending character: leave */
-      c[0] = '0';
-      return -1;
+      c[0] = '\0';
+      return 0;
+   }
+
+   mbSz = mblen(iinfo->preRead, iinfo->preReadSz);
+   if (mbSz > 0) {
+     mbSz = mbtowc(c, iinfo->preRead, mbSz);
+     if (mbSz > 0) {
+       iinfo->preReadSz -= mbSz;
+       memmove(iinfo->preRead, iinfo->preRead + mbSz, iinfo->preReadSz);
+     } else {
+       iinfo->preReadSz = 0;
+       return -1;
+     }
    }
-       
-   c[0] = iinfo->preRead[0];
-   if (iinfo->preReadSz-- > 0)
-      memmove(iinfo->preRead, iinfo->preRead+1, iinfo->preReadSz);
 
    return 1;
 }
@@ -200,7 +304,7 @@ elTclRead(ClientData data, int mask)
    /* we know that there is at least one character (since we're
     * called from a select() but there might be more than one. Thus we
     * try to compute the actual number and we keep on the safe side in
-    * the case it is not possible to get that number */ 
+    * the case it is not possible to get that number */
 
 #ifdef FIONREAD
    if (ioctl(0, FIONREAD, &nRead)) {
@@ -244,8 +348,8 @@ elTclGetWindowSize(int fd, int *lins, int *cols)
 	    *cols = ws.ws_col;
 	 if (ws.ws_row && lins)
 	    *lins = ws.ws_row;
+	 return 0;
       }
-      return 0;
    }
 #endif
 #ifdef TIOCGSIZE
@@ -256,8 +360,8 @@ elTclGetWindowSize(int fd, int *lins, int *cols)
 	    *cols = ts.ts_cols;
 	 if (ts.ts_lines && lins)
 	    *lins = ts.ts_lines;
+	 return 0;
       }
-      return 0;
    }
 #endif
 
